from napoleontoolbox.forecasting import forecasting_utility
import xgboost as xgb
from sklearn.model_selection import RandomizedSearchCV

class XGBModel():
    def __init__(self):
       self.model = xgb.XGBRegressor()


    def calibrate(self, X, y):#, method = 'standard'):
        X = X.fillna(0)
        y = y.fillna(0)
        params = {
            "colsample_bytree": [0.7, 0.3],
            "gamma": [0, 0.5],
            "learning_rate": [0.03, 0.3],  # default 0.1
            "max_depth": [4],  # default 3
            "n_estimators": [120],  # default 100
            "subsample": [0.6, 0.4]
        }
        cv = 2
        n_iter = 50
        # cv=3
        # n_iter = 200
        search = RandomizedSearchCV(self.model, param_distributions=params, random_state=42, n_iter=n_iter, cv=cv,
                                    verbose=1, n_jobs=1, return_train_score=True)
        search.fit(X, y)
        best_params = search.best_params_
        self.model = xgb.XGBRegressor(colsample_bytree = best_params['colsample_bytree'], gamma = best_params['gamma'], learning_rate = best_params['learning_rate'],
                                      max_depth = best_params['max_depth'], n_estimators = best_params['n_estimators'], subsample = best_params['subsample'])

    def fit(self, X_train, y_train, X_val, y_val):#, method = 'standard'):
        X_train = X_train.fillna(0)
        y_val = y_val.fillna(0)
        self.model.fit(X_train, y_train, eval_set=[(X_val, y_val)], eval_metric=['rmse', 'logloss'],
                           early_stopping_rounds=5, verbose=0)


    def predict(self, X_test):#, method = 'standard'):
        X_test = X_test.fillna(0)
        y_pred = self.model.predict(X_test)
        return y_pred


    def get_features_importance(self, features_names):#, method = 'standard'):
        pass
        # run_importances = {}
        # if method == 'lgbm':
        #     for (name, imp) in zip(features_names, self.model.feature_importances_):
        #         run_importances[name] = imp
        # if method == 'standard':
        #     for (name, imp) in zip(features_names, self.model.coef_):
        #         run_importances[name] = imp
        # return run_importances



from napoleontoolbox.forecasting import forecasting_utility
from sklearn.linear_model import Lasso
from sklearn.model_selection import cross_val_score, GridSearchCV, KFold, RandomizedSearchCV, train_test_split


class LassoModel():

    def __init__(self):#, method = 'standard'):
        alpha = 0.001
        self.model = Lasso(alpha=alpha, fit_intercept=False, max_iter=5000)        # """ Initialize shape of target. """

    def calibrate(self, X, y):#, method = 'standard'):

        param_grid = {
            'alpha': [0, 0.01, 0.1,0.5,1],
        }
        X = X.fillna(0)
        y = y.fillna(0)
        lasso_grid = GridSearchCV(self.model, param_grid, cv=2)
        lasso_grid.fit(X, y)
        print('Best parameters found by grid search are:', lasso_grid.best_params_)
        best_params = lasso_grid.best_params_
        self.model = Lasso(alpha=best_params['alpha'], fit_intercept=False, max_iter=5000)


    def fit(self, X_train, y_train, X_val, y_val):#, method = 'standard'):
        X_train = X_train.fillna(0)
        y_train = y_train.fillna(0)
        print(X_train, y_train)
        self.model.fit(X_train, y_train)


    def predict(self, X_test):#, method = 'standard'):
        X_test = X_test.fillna(0)
        y_pred = self.model.predict(X_test)
        print(y_pred)
        return y_pred


    def get_features_importance(self, features_names):#, method = 'standard'):
        pass
        # run_importances = {}
        # if method == 'lgbm':
        #     for (name, imp) in zip(features_names, self.model.feature_importances_):
        #         run_importances[name] = imp
        # if method == 'standard':
        #     for (name, imp) in zip(features_names, self.model.coef_):
        #         run_importances[name] = imp
        # return run_importances



import lightgbm as lgb

import numpy as np

from scipy.stats import uniform, randint

from sklearn.datasets import load_breast_cancer, load_diabetes, load_wine
from sklearn.metrics import auc, accuracy_score, confusion_matrix, mean_squared_error
from sklearn.model_selection import cross_val_score, GridSearchCV, KFold, RandomizedSearchCV, train_test_split
from sklearn.linear_model import Lasso
import xgboost as xgb
from sklearn import datasets, linear_model

from napoleontoolbox.forecasting import linear_model_wrapper
from napoleontoolbox.forecasting import lasso_model_wrapper
from napoleontoolbox.forecasting import lgbm_model_wrapper
from napoleontoolbox.forecasting import xgb_model_wrapper
from napoleontoolbox.forecasting import standard_mean_model_wrapper
from napoleontoolbox.forecasting import deterministic_sharpe_optim_model_wrapper
from napoleontoolbox.forecasting import to_weighted_mean_model_wrapper
from napoleontoolbox.forecasting import max_occurence_model_wrapper
from napoleontoolbox.forecasting import erc_allocation_model_wrapper
from napoleontoolbox.forecasting import mvp_allocation_model_wrapper
from napoleontoolbox.forecasting import mvp_uc_allocation_model_wrapper
from napoleontoolbox.forecasting import ivp_allocation_model_wrapper
from napoleontoolbox.forecasting import hrp_allocation_model_wrapper

from abc import ABC, abstractmethod


def instantiate_model(method):
    if method == 'standard':
        return linear_model_wrapper.LinearModel()
    if method == 'lasso':
        return lasso_model_wrapper.LassoModel()
#        alpha = 0.1
#        self.model = Lasso(alpha=alpha, fit_intercept=False, max_iter=5000)
    if method == 'lgbm':
        return lgbm_model_wrapper.LGBMModel()
    if method == 'xgb':
        return xgb_model_wrapper.XGBModel()
    if method == 'standard_mean':
        return standard_mean_model_wrapper.MeanModel()
    if method == 'to_weighted_mean':
        return to_weighted_mean_model_wrapper.WeightedMeanModel()
    if method == 'deterministic_sharpe_optim':
        return deterministic_sharpe_optim_model_wrapper.DeterministicSharpeOptimModel()
    if method == 'max_occurence':
        return max_occurence_model_wrapper.MaxOccurenceModel()
    if method == 'erc_allocation_model':
        return erc_allocation_model_wrapper.ERC_Allocation_Model()
    if method == 'mvp_allocation_model':
        return mvp_allocation_model_wrapper.MVP_Allocation_Model()
    if method == 'mvp_uc_allocation_model':
        return mvp_uc_allocation_model_wrapper.MVP_uc_Allocation_Model()
    if method == 'ivp_allocation_model':
        return ivp_allocation_model_wrapper.IVP_Allocation_Model()
    if method == 'hrp_allocation_model':
        return hrp_allocation_model_wrapper.HRP_Allocation_Model()



def report_best_scores(results, n_top=3):
    for i in range(1, n_top + 1):
        candidates = np.flatnonzero(results['rank_test_score'] == i)
        for candidate in candidates:
            print("Model with rank: {0}".format(i))
            print("Mean validation score: {0:.3f} (std: {1:.3f})".format(
                  results['mean_test_score'][candidate],
                  results['std_test_score'][candidate]))
            print("Parameters: {0}".format(results['params'][candidate]))
            print("")

class AbstractForecasterWrapper(ABC):
    def __init__(self):
        pass
    @abstractmethod
    def calibrate(self, X, y):#, method = 'standard'):
        pass

    @abstractmethod
    def fit(self, X_train, y_train, X_val, y_val):#, method = 'standard'):
        pass

    @abstractmethod
    def predict(self, X_test):#, method = 'standard'):
        pass

    @abstractmethod
    def get_features_importance(self, features_names):#, method = 'standard'):
        pass



#!/usr/bin/env python3
# coding: utf-8

import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import datetime
from napoleontoolbox.analyzer import market

def rand_color_palette(N):
    col = []
    colors = sns.mpl_palette('Set1', 9)
    for j in range(N):
        if j == 9:
            colors = sns.mpl_palette('Set3', 12)

        elif j == 21:
            colors = sns.mpl_palette('Set2', 8)

        elif j == 29:
            colors = list(sns.crayons.keys())

        i = np.random.randint(0, high=len(colors))
        if j >= 29:
            col += [sns.crayon_palette(colors.pop(i))]

        else:
            col += [colors.pop(i)]

    return col


palette = rand_color_palette(17)


def display_results(df, *args, w_mat=None, title='', palette=None, figsize=(16, 16),strats=None):
    N = df.columns.size - len(args)
    if palette is None:
        palette = rand_color_palette(N)

    if w_mat is None:
        f, ax = plt.subplots(1, 1, figsize=figsize)

    else:
        n = sum(w_mat.sum(axis=1) == 0.)
        f, (ax, ax2) = plt.subplots(2, 1, figsize=figsize)
        ax4 = f.add_subplot(2, 2, 4)
        ax3 = f.add_subplot(2, 2, 3, sharey=ax2)
        w_mat.iloc[n:].mean(axis=0).plot(
            kind='pie',
            ax=ax4,
            title='Mean of weights allocation',
            colors=palette
        )
        ax4.set_ylabel('weights in %', fontsize=12, y=0.5, x=-0.5, rotation=0)
        ax2.set_xticks([])
        w_mat.iloc[n:].plot(
            ax=ax3,
            kind='area',
            stacked=True,
            title='Weights allocation',
            color=palette
        )

    ma = market.MarketAnalyzer(df)
    ma.display_kpi()
    ma.plot_perf(logy=True, title=title, y=strats, ax=ax, color=palette, show=False)
    ma.plot_perf(logy=True, title=title, y=list(args), ax=ax, c='k', lw=2)


def display_weight_overtime(w_mat=None, palette=None, figsize=(16, 16)):
    N = w_mat.columns.size
    if palette is None:
        palette = rand_color_palette(N)
    f, ax = plt.subplots(1, 1, figsize=figsize)
    n = sum(w_mat.sum(axis=1) == 0.)
    w_mat.iloc[n:].plot(
        ax=ax,
        kind='area',
        stacked=True,
        title='Features importance',
        color=palette
    )

def display_weight_camember(w_mat=None, palette=None, figsize=(16, 16)):
    N = w_mat.columns.size
    if palette is None:
        palette = rand_color_palette(N)
    f, ax = plt.subplots(1, 1, figsize=figsize)
    n = sum(w_mat.sum(axis=1) == 0.)
    w_mat.iloc[n:].mean(axis=0).plot(
        kind='pie',
        ax=ax,
        title='Mean of features group importance',
        colors=palette
    )


def perfligne(df, perf):
    init = df.index[1].year
    initm = df.index[1].month
    fin = df.index[-1].year
    finm = df.index[-1].month
    A = np.arange(init, fin + 1, 1)
    S = np.zeros([A.shape[0], 13])  # * np.nan
    end_month = df[perf].resample('1M').last()
    end_month_return = (end_month / end_month.shift(1) - 1).fillna(0)
    # Annee zero
    S[0, initm - 1] = end_month.iloc[0] / df[perf][0] - 1
    if initm < 12:
        for i in range(initm, 12):
            S[0, i] = end_month_return[pd.Series(pd.date_range('1/1/{}'.format(init), freq='M', periods=12))[i]]
    S[0, 12] = end_month.loc[datetime.datetime(init, 12, 31)] / df[perf][0] - 1
    # Années 1 .. (n-1)
    for k in range(1, A.shape[0] - 1):
        year = A[k]
        for i in range(12):
            S[k, i] = end_month_return[pd.Series(pd.date_range('1/1/{}'.format(year), freq='M', periods=12))[i]]
        S[k, 12] = end_month.loc[datetime.datetime(year, 12, 31)] / end_month.loc[
            datetime.datetime(year - 1, 12, 31)] - 1
    # Année n
    year = fin
    if finm == 1:
        S[A.shape[0] - 1, 0] = y[perf].values[-1] / end_month.loc[datetime.datetime(year - 1, 12, 31)] - 1
    else:
        for i in range(finm):
            S[A.shape[0] - 1, i] = end_month_return[
                pd.Series(pd.date_range('1/1/{}'.format(year), freq='M', periods=12))[i]]
    S[A.shape[0] - 1, 12] = end_month.iloc[-1] / end_month.loc[datetime.datetime(year - 1, 12, 31)] - 1
    return S


def display_perf_table(df, perf, vmin1=None, vmax1=None, vmin2=None, vmax2=None):
    """
    Display a table of monthly and yearly performances of column "perf" in dataframe df
    Setup gradient colors :
    vmin1, vmax1 range monthly variations
    vmin2,vmax2 range yearly variations
    """
    df.index = pd.to_datetime(df.index)
    init = df.index[1]
    A = np.arange(init.year, datetime.datetime.today().year + 1, 1)

    dfmonth = pd.DataFrame(index=A,
                           columns=["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec",
                                    "Year"],
                           data=perfligne(df, perf))
    dfmonth = dfmonth.style.background_gradient(cmap=sns.diverging_palette(17, 170, n=5, l=70, as_cmap=True), axis=None,
                                                vmin=vmin1, vmax=vmax1,
                                                subset=["Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep",
                                                        "Oct", "Nov", "Dec", ]).background_gradient(
        cmap=sns.diverging_palette(17, 170, n=5, l=70, as_cmap=True), axis=0, vmin=vmin2, vmax=vmax2,
        subset=["Year"])
    dfmonth = dfmonth.format({
        'Jan': '{:.2%}',
        'Feb': '{:.2%}',
        'Mar': '{:.2%}',
        'Apr': '{:.2%}',
        'May': '{:.2%}',
        'Jun': '{:.2%}',
        'Jul': '{:.2%}',
        'Aug': '{:.2%}',
        'Sep': '{:.2%}',
        'Oct': '{:.2%}',
        'Nov': '{:.2%}',
        'Dec': '{:.2%}',
        'Year': '{:.2%}',
    })
    return dfmonth